@@ -40,13 +40,31 @@ DEFAULT_LAYOUT = "layouts/default.html"
 
 HTML_REDIRECT_TEMPLATE = """<meta http-equiv="Refresh" content="0; url='{url}'" />"""
 
+
+__slots__ = [ "pages_data__", "resources__" ]
+
+pages_data__ = {}
+resources__ = {}
+
+
 def print_info(message):
     print("- %s" % message)
 
 cached = {}
     
-class Generator(object):
+def _clean_hashfile_key(src_file) -> str:
+  """
+  create a name for the hash key map by removing extension
+  Returns str
+  """
+  if not src_file.endswith((".html", ".md")):
+    src_file += ".html"
+  short_name, _ = os.path.splitext(src_file)
+  return short_name
 
+
+class Generator(object):
+    
     default_page_meta = {
         "title": "",            # The title of the page
         "markup": None,         # The markup to use. ie: md | html (default)
@@ -68,7 +86,7 @@ class Generator(object):
     }
     tpl_env = None
     _templates = {}
-    _pages_meta = {}
+    #pages_data__ = {}
 
     def __init__(self, root_dir, options={}):
         """
@@ -116,18 +134,16 @@ class Generator(object):
         self.base_url = self.site_config.get("base_url")
         self.static_url = self.site_config.get("static_url")
         self._context_root = "./"
-        self._data = {}
-        self._load_data()
+        # self._data = {}
+        self._load_data_resources()
 
-        # Map the pages so it can be easily resolve, ie: index -> index.html, hello/world -> hello/world.html
-        self.pages_short_mapper = {}
         self.print_info = False
 
         self._init_jinja({
             "get_page": self._get_page,
             "get_page_url": self._get_page_url,
             "get_static_url": self._get_static_url,
-            "get_data": self._get_data,
+            "get_resources": self._get_resouces,
             "get_static_bundle": self._get_page_static_bundle_url,
             "get_site": self._get_site,
             "format_date": lambda dt, format="MM/DD/YYYY h:mm a": arrow.get(dt).format(format),
@@ -135,8 +151,7 @@ class Generator(object):
             "site": self.site_config,
             "__info": self._load_info(),
         })
-
-        
+ 
     def _load_info(self):
         """ Global variables """
         now = datetime.datetime.now()
@@ -170,7 +185,7 @@ class Generator(object):
         self.tpl_env.add_extension('jinja2.ext.do')
         self.tpl_env.globals.update(global_context)
 
-    def _get_data(self, path):
+    def _get_resouces(self, path):
         """
         Get the data from /data dir
 
@@ -178,38 +193,35 @@ class Generator(object):
             path:dotNotationStr
         Returns: mixed
         """
-        return self._data.get(path)
+        return resources__.get(path)
 
     def _get_page_meta_data(self, page):
-        """
-        Cache the page meta from the frontmatter and assign new keys
-        The cache data will be used to build links or other properties
-        """
-        page = self.pages_short_mapper.get(page)
-        meta = self._pages_meta.get(page)
-        if not meta:
-            src_file = os.path.join(self.pages_dir, page)
-            with open(src_file) as f:
-                _, _ext = os.path.splitext(src_file)
-                markup = _ext.replace(".", "")
-                _meta, _ = frontmatter.parse(f.read())
-                meta = self.default_page_meta.copy()
-                meta["meta"].update(self.config.get("site.meta", {}))
-                meta.update(_meta)
-                dest_file, url = self._get_dest_file_and_url(page, meta)
-                meta["url"] = url
-                meta["filepath"] = dest_file
-                if meta.get("markup") is None:
-                    meta["markup"] = markup
-                self._pages_meta[page] = lib.dictdot(meta)
-        return meta
+        """ A shortcut to retrieve the page meta data """
+        hashed_name = _clean_hashfile_key(page)
+        return pages_data__.get(hashed_name)
 
-    def _get_page_content(self, page):
-        """ Get the page content without the frontmatter """
+    def _get_file_data(self, page) -> dict:
+        """
+        Retrieve the file data that's in the front matter
+
+        Returns: dict
+        """
+        meta = {}
         src_file = os.path.join(self.pages_dir, page)
         with open(src_file) as f:
-            _meta, content = frontmatter.parse(f.read())
-            return content
+            _, _ext = os.path.splitext(src_file)
+            markup = _ext.replace(".", "")
+            _meta, _content = frontmatter.parse(f.read())
+            meta = self.default_page_meta.copy()
+            meta["meta"].update(self.config.get("site.meta", {}))
+            meta.update(_meta)
+            dest_file, url = self._get_dest_file_and_url(page, meta)
+            meta["url"] = url
+            meta["filepath"] = dest_file
+            meta["is_generator"] = "_generator" in meta
+            meta["markup"] = meta.get("markup") or markup
+            meta["content"] = _content
+            return lib.dictdot(meta)
 
     def _get_page(self, page_dot_path:str):
         """
@@ -295,25 +307,305 @@ class Generator(object):
         url = "/" + dest_file.replace("index.html", "")
         return dest_file, url
 
-    def _load_data(self):
+    def _load_data_resources(self):
+        global resources__
         data = {}
-        # Load data from the data directory
+
+        # load data from the data directory
         for root, _, files in os.walk(self.data_dir):
             for fname in files:
-                if fname.endswith((".json",)):
-                    name = fname.replace(".json", "")
+                if fname.endswith((".json", ".yml")):
+                    name = fname.replace(".json", "").replace(".yml", "")
                     fname = os.path.join(root, fname)
                     if os.path.isfile(fname):
-                        with open(fname) as f:
-                            _ = json.load(f)
-                            if isinstance(_, dict):
-                                _ = lib.dictdot(_)
-                            data[name] = _
-        self._data = lib.dictdot(data)
+                        m = lib.load_yaml_file if fname.endswith(".yml") else lib.load_json_file
+                        data[name] = m(fname)
+
+        # load the data from config.remote_resources
+        """
+        format:
+            name:str
+            request:dict
+                url:str
+                method:str
+                headers:dict|None
+                data:dict|None
+                params:dict|None
+                data_path:str - the path to resolve the data
+            cached:bool - To use a cached version when getting rebuilt. To improve speed
+            saved:bool - to save as file. Takes precedence over cached, as the file will exist
+            disabled:bool - to disabled the request
+        """
+
+        if remote_resources := self.config.get("remote_resources"):
+            for d in remote_resources:
+                if name := d.get("name"):
+                    if not d.get("request"):
+                        raise Exception("BUILD ERROR: remote resource '%s' missing request object" % name)
+                    if d.get("disabled") is True:
+                        continue
+                    if d.get("cached") is True and data.get(name):
+                        continue
+
+                    if _data := lib.make_http_requests(**d.get("request")):
+                        # save to data
+                        if d.get("saved") is True:
+                            save_file = "%s/%s.json" % (self.data_dir.rstrip("/"), name)
+                            if not os.path.exists(save_file):
+                                lib.write_file(dest=save_file, contents=lib.json_dumps(_data))
+                        data[name] = _data
+
+        resources__ = lib.dictdot(data)
 
     def _get_site(self, prop):
         return self.site_config.get(prop)
 
+    def _build_page(self, filepath) -> list:
+        
+        """ To build from filepath, relative to pages_dir """
+        filename = filepath.split("/")[-1]
+        # If filename starts with _ (underscore) or . (dot) do not build
+        if not filename.startswith(("_", ".")) and (filename.endswith(PAGE_FORMAT)):
+            # The default context for the page
+            _default_page = {
+                "build_dir": self.build_dir,
+                "filepath": None,
+                "context": {"page": {}, "this": {}},
+                "content": "",
+                "markup": None,
+                "template": None,
+                "layout": None or self.layout
+            }
+
+            # only non generator will contain the data
+            if mdata := self._get_page_meta_data(filename):
+                #!== regular pages
+                if not mdata.get("is_generator"):
+                    return [{
+                        "build_dir": self.build_dir,
+                        "filepath": mdata.get("filepath"),
+                        "context": {"page": mdata, "this": mdata},
+                        "content": mdata.get("content"),
+                        "markup": mdata.get("markup"),
+                        "template": mdata.get("template"),
+                        "layout": mdata.get("layout") or self.layout
+                    }]
+
+                else:
+                    """
+                    #!== Generator 
+                    Generators generates pages from data sources.
+                    A data source can be from the data folder, or from a `data_requests` in ssg.yml
+
+                    Properties:
+                        - resources: somename_from_dataDir_or_remoteResources
+                        - type: single|paged
+
+
+                    Example: 
+                    Having the data `data/articles.json` -> [{title, slug, content, ...},...]
+                    and the page `pages/article_generator.html`
+
+                    ``` pages/article_generator.html
+                    ---
+                    _generator:
+                        - resources: articles
+                        - type: single
+                        - slug: article/{slug}
+                    ---
+                        <h1>{{ this.title }}</h1>
+                        <main>{{ this.content }}</main>
+                    ```
+                    """
+                    lpages = []
+                    # We want these back in meta in they exists in the data
+                    special_meta = ["title", "permalink", "description"]                    
+                    _generator = mdata.get("_generator")
+                    _gen_type = _generator.get("type", "single") # The generator type: single|list
+                    _resources_name = _generator.get("resources")
+                    rdata = resources__.get(_resources_name)
+                    # expecting data to be a list to iterate over
+                    if not isinstance(rdata, list):
+                        raise Exception("BUILD ERROR: Generator data type must be of List|Array for '%s'" % _resources_name)
+
+                    # SINGLE
+                    # generates single pages off of the data
+                    if not _gen_type or _gen_type.upper() == "SINGLE":
+                        
+                        for data in rdata:
+                            dmeta = copy.deepcopy(mdata)
+
+                            for _ in special_meta:
+                                if _ in data:
+                                    dmeta[_] = data.get(_)
+
+                            # If generator has the permalink, it will substitute if
+                            # Slug in the generator must have token from the data
+                            # to generate the slug
+                            if "permalink" in _generator:
+                                dmeta["permalink"] = lib.slugify_with_slash(_generator.get("permalink").format(**data))
+
+                            # Slug is required
+                            if "permalink" not in dmeta:
+                                print("WARNING: Skipping page because it's missing `permalink`")
+                                continue
+
+                            permalink = dmeta.get("permalink")
+                            dmeta["url"] = permalink
+                            this = {
+                                **dmeta,
+                                "data": data,
+                            }
+                            lpages.append({
+                                **copy.deepcopy(_default_page),
+                                "filepath": permalink,
+                                "context": {
+                                    "page": this,
+                                    "this": this
+                                },
+                                "content": mdata.get("content"),
+                                "markup": mdata.get("markup"),
+                                "template": mdata.get("template"),
+                                "layout": mdata.get("layout") or self.layout
+                            })
+   
+                    # LIST
+                    # create a paginated with [Prev] [Next]
+                    # Each page will contain a chunk of the data
+                    # will split the content into multi pages
+                    if _gen_type.upper() == "LIST":
+                        raise NotImplementedError()
+
+                        per_page = int(_generator.get(
+                            "per_page", self.site_config.get("pagination.per_page", 10)))
+                        left_edge = int(_generator.get(
+                            "left_edge", self.site_config.get("pagination.left_edge", 2)))
+                        left_current = int(_generator.get(
+                            "left_edge", self.site_config.get("pagination.left_current", 3)))
+                        right_current = int(_generator.get(
+                            "right_current", self.site_config.get("pagination.right_current", 4)))
+                        right_edge = int(_generator.get(
+                            "right_edge", self.site_config.get("pagination.right_edge", 2)))
+                        padding = _generator.get("padding")
+                        slug = _generator.get("slug")
+                        limit = _generator.get("limit")
+
+                        if "limit" in _generator:
+                            data = data[:int(limit)]
+                        data_chunks = lib.chunk_list(data, per_page)
+                        len_data = len(data)
+                        total_pages = len(data_chunks)
+
+                        for i, d in enumerate(data_chunks):
+                            dmeta = copy.deepcopy(mdata)
+                            page = copy.deepcopy(_default_page)
+
+                            page_num = i + 1
+                            _paginator = Paginator([],
+                                                total=len_data,
+                                                page=page_num,
+                                                per_page=per_page,
+                                                padding=padding,
+                                                left_edge=left_edge,
+                                                right_edge=right_edge,
+                                                left_current=left_current,
+                                                right_current=right_current)
+                            _paginator.slug = slug
+                            _paginator.index_slug = _generator.get("index_slug")
+
+                            current_page = i + 1
+                            prev_page = current_page - 1
+                            has_prev_page = prev_page < 0
+                            next_page = current_page + 1
+                            has_next_page = next_page < total_pages
+
+                            pagination = {
+                                "current_page": current_page,
+                                "prev_page": prev_page,
+                                "has_prev_page": has_prev_page,
+                                "next_page": next_page,
+                                "has_next_page": has_next_page
+                            }
+
+                            dmeta = copy.deepcopy(mdata)
+
+                            for _ in special_meta:
+                                if _ in data:
+                                    dmeta[_] = data.get(_)
+
+                            # If generator has the permalink, it will substitute if
+                            # Slug in the generator must have token from the data
+                            # to generate the slug
+                            if "permalink" in _generator:
+                                dmeta["permalink"] = lib.slugify_with_slash(_generator.get("permalink").format(**data))
+
+                            # Slug is required
+                            if "permalink" not in dmeta:
+                                print("WARNING: Skipping page because it's missing `permalink`")
+                                continue
+
+                            permalink = dmeta.get("permalink")
+                            dmeta["url"] = permalink
+                            this = {
+                                **dmeta,
+                                "data": data,
+                            }
+                            lpages.append({
+                                **copy.deepcopy(_default_page),
+                                "filepath": permalink,
+                                "context": {
+                                    "page": this,
+                                    "this": this
+                                },
+                                "content": mdata.get("content"),
+                                "markup": mdata.get("markup"),
+                                "template": mdata.get("template"),
+                                "layout": mdata.get("layout") or self.layout,
+                                "pagination": {
+                                    "has_prev": ""
+                                }
+                            })
+
+
+
+                            _slug = slug.format(**{"page_num": page_num})
+                            dmeta["url"] = _slug
+                            dmeta["context"] = d
+                            dmeta["paginator"] = _paginator
+
+
+                            permalink = dmeta.get("permalink")
+                            dmeta["url"] = permalink
+                            this = {
+                                **dmeta,
+                                "data": data,
+                            }
+
+
+
+
+
+                            page.update({
+                                "filepath": _slug,
+                                "context": {"page": dmeta}
+                            })
+                            self.create_page(**page)
+
+                            # First page need to generate the index
+                            if i == 0 and _generator.get("index_slug"):
+                                page["filepath"] = _generator.get("index_slug")
+                                self.create_page(**page)
+
+                    return lpages
+        
+        return []
+
+    def _reset_page_context_assets(self, context):
+        context["page"]["__assets"] = {
+            "js": [],
+            "css": []
+        }
+
     def clean_build_dir(self):
         if os.path.isdir(self.build_dir):
             shutil.rmtree(self.build_dir)
@@ -323,15 +615,17 @@ class Generator(object):
         """ Build static files """
         if not os.path.isdir(self.build_static_dir):
             os.makedirs(self.build_static_dir)
+        if not os.path.isdir(self.build_static_gen_dir):
+            os.makedirs(self.build_static_gen_dir)
+            
         if (self.print_info):
             print_info('copying static dir to build folder...')
         copy_tree(self.static_dir, self.build_static_dir)
 
-    def build_pages(self):
+    def build_pages(self) -> int:
         """Iterate over the pages_dir and build the pages """
         src_files = []
-        self._pages_meta = {}
- 
+
         # Aggregate all the files
         for root, _, files in os.walk(self.pages_dir):
             if (self.print_info):
@@ -340,138 +634,22 @@ class Generator(object):
             base_dir = root.replace(self.pages_dir, "").lstrip("/")
             if not base_dir.startswith("_"):
                 for f in files:
-                    fname, _ext = os.path.splitext(f)
-                    self.pages_short_mapper.update({fname: f, f: f})
-                    markup = _ext.replace(".", "")
                     src_file = os.path.join(base_dir, f)
                     src_files.append(src_file)
+                    _file_meta = self._get_file_data(src_file)
+                    hashed_name = _clean_hashfile_key(src_file)
+                    pages_data__[hashed_name] = _file_meta
+                    
         # Build pages
         if (self.print_info):
             print_info('initiating page building...')
-        [self._build_page(src_file) for src_file in src_files]
+        
+        _pages = [p for s in src_files for p in self._build_page(s) if p]
 
-    def _build_page(self, filepath):
-        """ To build from filepath, relative to pages_dir """
-        filename = filepath.split("/")[-1]
-        # If filename starts with _ (underscore) or . (dot) do not build
-        if not filename.startswith(("_", ".")) and (filename.endswith(PAGE_FORMAT)):
-            meta = self._get_page_meta_data(filepath)
-            content = self._get_page_content(filepath)
 
-            # The default context for the page
-            _default_page = {
-                "build_dir": self.build_dir,
-                "filepath": meta["filepath"],
-                "context": {"page": meta},
-                "content": content,
-                "markup": meta.get("markup"),
-                "template": meta.get("template"),
-                "layout": meta.get("layout") or self.layout
-            }
-
-            # GENERATOR
-            # Allows to generate
-            _generator = meta.get("_generator")
-            if _generator:
-                raise NotImplementedError('GENERATOR IS NOT PROPERLY IMPLEMENTED YET')
-                exit()
-                data = self._data.get(_generator.get("data_source"))
-
-                # We want these back in meta in they exists in the data
-                special_meta = ["title", "slug", "description"]
-
-                # SINGLE
-                if _generator.get("type") == "single":
-                    for d in data:
-                        dmeta = copy.deepcopy(meta)
-                        page = copy.deepcopy(_default_page)
-                        for _ in special_meta:
-                            if _ in d:
-                                dmeta[_] = d.get(_)
-
-                        # If generator has the slug, it will substitute if
-                        # Slug in the generator must have token from the data
-                        # to generate the slug
-                        if "slug" in _generator:
-                            dmeta["slug"] = _generator.get("slug").format(**d)
-
-                        # Slug is required
-                        if "slug" not in dmeta:
-                            print(
-                                "WARNING: Skipping page because it's missing `slug`")
-                            continue
-                        slug = dmeta.get("slug")
-                        dmeta["url"] = slug
-                        dmeta["context"] = d
-
-                        page.update({
-                            "filepath": slug,
-                            "context": {"page": dmeta}
-                        })
-                        self.create_page(**page)
-
-                if _generator.get("type") == "pagination":
-
-                    per_page = int(_generator.get(
-                        "per_page", self.site_config.get("pagination.per_page", 10)))
-                    left_edge = int(_generator.get(
-                        "left_edge", self.site_config.get("pagination.left_edge", 2)))
-                    left_current = int(_generator.get(
-                        "left_edge", self.site_config.get("pagination.left_current", 3)))
-                    right_current = int(_generator.get(
-                        "right_current", self.site_config.get("pagination.right_current", 4)))
-                    right_edge = int(_generator.get(
-                        "right_edge", self.site_config.get("pagination.right_edge", 2)))
-                    padding = _generator.get("padding")
-                    slug = _generator.get("slug")
-                    limit = _generator.get("limit")
-
-                    if "limit" in _generator:
-                        data = data[:int(limit)]
-                    data_chunks = lib.chunk_list(data, per_page)
-                    len_data = len(data)
-
-                    for i, d in enumerate(data_chunks):
-                        dmeta = copy.deepcopy(meta)
-                        page = copy.deepcopy(_default_page)
-
-                        page_num = i + 1
-                        _paginator = Paginator([],
-                                               total=len_data,
-                                               page=page_num,
-                                               per_page=per_page,
-                                               padding=padding,
-                                               left_edge=left_edge,
-                                               right_edge=right_edge,
-                                               left_current=left_current,
-                                               right_current=right_current)
-                        _paginator.slug = slug
-                        _paginator.index_slug = _generator.get("index_slug")
-
-                        _slug = slug.format(**{"page_num": page_num})
-                        dmeta["url"] = _slug
-                        dmeta["context"] = d
-                        dmeta["paginator"] = _paginator
-                        page.update({
-                            "filepath": _slug,
-                            "context": {"page": dmeta}
-                        })
-                        self.create_page(**page)
-
-                        # First page need to generate the index
-                        if i == 0 and _generator.get("index_slug"):
-                            page["filepath"] = _generator.get("index_slug")
-                            self.create_page(**page)
-
-            # NORMAL PAGE
-            else:
-                self.create_page(**_default_page)
-
-    def _reset_page_context_assets(self, context):
-        context["page"]["__assets"] = {
-            "js": [],
-            "css": []
-        }
+        for p in _pages:
+            self.create_page(**p)
+        return len(_pages)
 
     def create_page(self, build_dir, filepath, context={}, content=None, template=None, markup=None, layout=None):
         """
@@ -514,14 +692,13 @@ class Generator(object):
         if not os.path.isdir(dest_dir):
             os.makedirs(dest_dir)
 
-        
         if "page" not in _context:
             _context["page"] = self.default_page_meta.copy()
         if "url" not in _context["page"]:
             _context["page"]["url"] = "/" + \
                 filepath.lstrip("/").replace("index.html", "")
+            print("NOT HERE", filepath, _context["page"]["url"])
         self._reset_page_context_assets(_context)
-        _context["this"] = _context["page"]
 
         #! redirect
         if redirect := _context.get("page.redirect"):
@@ -556,16 +733,14 @@ class Generator(object):
                 
             # Create SFC Assets
             if is_sfc is True:
-                if not os.path.isdir(self.build_static_gen_dir):
-                    os.makedirs(self.build_static_gen_dir)
-                sfc_hash = lib.gen_hash()
+                sfc_uuid = lib.gen_uuid()
                 sfc_asset_filepath = slugify(filepath)
                 sfc_o = {"script": "js", "style": "css"}
                 for o in sfc_o:
 
                     if (sfc_c.get(o)):
                         _ff = os.path.join(self.build_static_gen_dir, "%s_%s.%s" % (
-                            sfc_asset_filepath, sfc_hash, sfc_o[o]))
+                            sfc_asset_filepath, sfc_uuid, sfc_o[o]))
                         _sff = _ff.replace(self.build_static_dir, '').lstrip("/")
                         _context["page"]["sfc_%s_path" % sfc_o[o]] = _sff
 
@@ -606,7 +781,7 @@ class Generator(object):
                 {% endfor %}
 
             """
-            name = "bundle_%s_%s" % (slugify(filepath), lib.gen_hash())
+            name = "bundle_%s_%s" % (slugify(filepath), lib.gen_uuid())
             filepath_ = os.path.join(self.build_static_gen_dir, name)
             bundlename_ = filepath_.replace(self.build_static_dir, '').lstrip("/")
         
@@ -635,10 +810,10 @@ class Generator(object):
         self._context_root = lib.context_root(_context.get("page.url"))
         lib.write_file(dest=dest_file, contents=tpl.render(**_context))
 
-    def build(self, print_info=False):
+    def build(self, print_info=False) -> int:
         self.print_info = print_info
         self.clean_build_dir()
         if not os.path.isdir(self.build_dir):
             os.makedirs(self.build_dir)
         self.build_static()
-        self.build_pages()
+        return self.build_pages()
